/*  MODULE     :  tokenizer.c
 *  DESCRIPTION:  This module performs various functions on a stream
 *  		  of tokens using the module stream.c.  
 */
 
#include <stdio.h>
#include "tokens.h"
#include "stream.h"

/**************************************************************/
/*  FUNCTION   : open_token_stream 
 *  INPUTS     : filename 
 *  OUTPUTS    : TOKEN_STREAM 
 *  DESCRIPTION: associates a filename to a token stream using 
 * 		 open_character_stream. Strings are extracted from
 * 		 character_stream and put in token format. 
 */ 
 
TOKEN_STREAM open_token_stream(filename)
char *filename;
{
   CHARACTER_STREAM stream; 
   TOKEN_STREAM     ts;
   TOKEN            word;

   stream = open_character_stream(filename);   
   ts.pos = 0;
   while (!is_end_of_character_stream(stream))
      {
	ts.token_array[ts.pos] = make_token(&stream); 
        ++ts.pos; 
      }
   ts.pos = 0; 
   return(ts); 
}

/**************************************************************/
/*  FUNCTION   : remove_comments 
 *  INPUTS     : character_stream 
 *  OUTPUTS    : none 
 *  DESCRIPTION: removes comments from scheme input 
 */ 

void remove_comments(stream)
CHARACTER_STREAM *stream;
{
   char character;

   while ((character != '\n')&&(!is_end_of_character_stream(*stream)))
      character = get_char(stream);
}

/**************************************************************/
/*  FUNCTION   : get_symbol_keyword 
 *  INPUTS     : character_stream 
 *  OUTPUTS    : TOKEN 
 *  DESCRIPTION: get the keyword "=>" from character stream 
 */ 

TOKEN get_symbol_keyword(stream) 
CHARACTER_STREAM *stream;
{
   TOKEN token; 
   char character;

   if ((character=get_char(stream))=='>')
      {
         strcpy(token.first_element,"keyword");
         strcpy(token.second_element,"=>");
      }
   else
      {
         token=error_token(character);
      }
   return(token);
}

/**************************************************************/
/*  FUNCTION   : get_alpha_numeric 
 *  INPUTS     : character_stream 
 *  OUTPUTS    : TOKEN 
 *  DESCRIPTION: get an alpha-numeric string from character stream.
 * 		 a function is called to determine if the string
 * 		 is an identifier or a keyword. 
 */ 

TOKEN get_alpha_numeric(stream)
CHARACTER_STREAM *stream;
{
   TOKEN token;
   char character;
   int temp_index = 0;
   char temp_string[MAX_TOKEN_SIZE];

   character = get_char(stream);
   while ((isdigit(character)||(isalpha(character)))
                &&(!is_end_of_character_stream(*stream)))
      {
        temp_string[temp_index++] = character;
        temp_string[temp_index] = '\0';
        character = get_char(stream);
      }  
   if ((symbol(character))||(is_end_of_character_stream(*stream))||
        (isspace(character)))
      {  
        /* alpha numeric constant terminated normally */
        strcpy(token.second_element,temp_string);
        push_char(character,stream);
      }  
   else
        token=error_token(character);

   return(token);
}

/**************************************************************/
/*  FUNCTION   : is_keyword 
 *  INPUTS     : string
 *  OUTPUTS    : boolean 
 *  DESCRIPTION: a string is passed to this function to determine
 *  		 if it is a keyword or an identifier.  The result
 * 		 is passed back as a boolean. 
 */ 

int is_keyword(string) 
char string[MAX_TOKEN_SIZE];
{
   if ((!strcmp(string,"=>"))||
	(!strcmp(string,"and"))|| 
	(!strcmp(string,"or"))||
	(!strcmp(string,"if"))||
	(!strcmp(string,"xor"))||
	(!strcmp(string,"lambda"))) 
	   return(TRUE);
   else 
	   return(FALSE);
}

/**************************************************************/
/*  FUNCTION   : get_string_constant  
 *  INPUTS     : character_stream 
 *  OUTPUTS    : TOKEN 
 *  DESCRIPTION: if the character " is read then this function is
 *  		 called to get the string constant. 
 */ 

TOKEN get_string_constant(stream)
CHARACTER_STREAM *stream;
{
   char character;
   TOKEN token;
   char temp_string[MAX_TOKEN_SIZE];
   int temp_index = 0;

   while ((character=get_char(stream)!='\"')&&
		(!is_end_of_character_stream(*stream)))
      {
        temp_string[temp_index++] = character;
        temp_string[temp_index] = '\0';
      }
   if (character == '\"')
      {
        strcpy(token.first_element,"string");
        strcpy(token.second_element,temp_string);
      }
   else
	token=error_token(character);

   return(token);
} 

/**************************************************************/
/*  FUNCTION   : get_numeric_constant 
 *  INPUTS     : character_stream 
 *  OUTPUTS    : TOKEN 
 *  DESCRIPTION: retrieve numeric constant from character stream. 
 */ 

TOKEN get_numeric_constant(stream)
CHARACTER_STREAM *stream;
{
   char character;
   TOKEN token;
   char temp_string[MAX_TOKEN_SIZE];
   int temp_index = 0;  

   character = get_char(stream); 
   while (isdigit(character)&&(!is_end_of_character_stream(*stream)))
      {
        temp_string[temp_index++] = character;
        temp_string[temp_index] = '\0';
        character = get_char(stream);
      }
   if ((symbol(character))||(is_end_of_character_stream(*stream))||
	(isspace(character)))
      {
	/* numeric constant terminated normally */
        push_char(character,stream);   
        strcpy(token.first_element,"numeric");
        strcpy(token.second_element,temp_string);
      }
   else
      {
	token=error_token(character);
      }
   
   return(token);
}

/**************************************************************/
/*  FUNCTION   : get_character_constant  
 *  INPUTS     : stream 
 *  OUTPUTS    : TOKEN 
 *  DESCRIPTION: retrieve a character constant from a character 
 *  		 stream.  Character constants are preceeded by the
 * 		 symbol '#'.  Valid characters are any printable
 * 		 characters. 
 */ 

TOKEN get_character_constant(stream)
CHARACTER_STREAM *stream;
{
   char character,next_character;
   TOKEN token;

   character=get_char(stream);
   next_character=get_char(stream);
   
   if (isspace(next_character)||symbol(next_character))
      {
      	strcpy(token.first_element,"character");
	token.second_element[0] = character;
	token.second_element[1] = '\0';
      }
   else
      {
        token=error_token(character);
        return(token);
      }
   push_char(next_character,stream);
   return(token);
}

/**************************************************************/
/*  FUNCTION   : error_token 
 *  INPUTS     : character 
 *  OUTPUTS    : TOKEN 
 *  DESCRIPTION: if an illegal character is encounted, it is 
 *   		 passed and assigned to a token of error type.  
 */ 

TOKEN error_token(character)
char character;
{
   TOKEN token;	

   strcpy(token.first_element,"error");
   token.second_element[0] = character;
   token.second_element[1] = '\0';

   return(token);
}

/**************************************************************/
/*  FUNCTION   : get_string 
 *  INPUTS     : stream 
 *  OUTPUTS    : TOKEN 
 *  DESCRIPTION: This function is called as a default case.  It  
 *		 entails any character that is not a symbol or 
 * 		 the identifier "=>" 
 */ 

TOKEN get_string(stream)
CHARACTER_STREAM *stream;
{
   char character;
   TOKEN token;

   character=get_char(stream);
   if isalpha(character) 
      {
        push_char(character,stream);
	token=get_alpha_numeric(stream);
        if (is_keyword(token.second_element))
	      strcpy(token.first_element,"keyword");
	else
	   if (token.first_element[0] == NULL)
	      strcpy(token.first_element,"identifier");
      }
   else
      if isdigit(character)
	{
	   push_char(character,stream);
           token = get_numeric_constant(stream);  
        }
   else
        if ((character == EOF) || (is_end_of_character_stream(stream)))
           {
              strcpy(token.first_element,"eof");
              token.second_element[0] = '\0';
           }
   else 
      {
        /* error token */ 
	token=error_token(character);
 
      }
   return(token);
}

/**************************************************************/
/*  FUNCTION   : make_token 
 *  INPUTS     : stream 
 *  OUTPUTS    : TOKEN 
 *  DESCRIPTION: This is the main function which determines what
 *   		 type of token needs to be read using a case     
 * 		 statement.  Then the appropriate function is
 *		 called to get a token.  This token is the returned. 
 */ 

TOKEN make_token(stream)
CHARACTER_STREAM *stream;
{
   char character;
   TOKEN token;
   int status = 0;

   token.first_element[0] = '\0';
   token.second_element[0] = '\0';
   while((isspace(character = get_char(stream)))&& 
		(!is_end_of_character_stream(stream))); 
   if (!isspace(character))
      {  
        switch(character)
           {
                case '(': strcpy(token.first_element,"lparen");
                          break; 
                case ')': strcpy(token.first_element,"rparen");
                          break;
                case '[': strcpy(token.first_element,"lsquare");
                          break;
                case ']': strcpy(token.first_element,"rsquare");
                          break;
                case '\'': strcpy(token.first_element,"quote");
                          break;
                case '`': strcpy(token.first_element,"bquote");
                          break;
		case ',': strcpy(token.first_element,"comma");
			  break;
          	case '=': token = get_symbol_keyword(stream); 
			  break; 
                	  /*  case for comments */
                case ';': remove_comments(stream);
			  break;
		case '\"':token = get_string_constant(stream);
			  break;
		case '#': token = get_character_constant(stream);
			  break;
                	  /*  case for strings */
                default :  
                        push_char(character,stream);
			token = get_string(stream);
                        break;
           }
      }
   else
      {
	strcpy(token.first_element,"eof");
        token.second_element[0] = '\0';
      }
   return(token); 
}

/**************************************************************/
/*  FUNCTION   : get_token 
 *  INPUTS     : token_stream 
 *  OUTPUTS    : TOKEN 
 *  DESCRIPTION: a token is retrieved from a token stream and  
 *   		 returned to the calling function 
 */ 

TOKEN get_token(token_stream)
TOKEN_STREAM *token_stream;
{
   TOKEN token;

   token = token_stream->token_array[token_stream->pos]; 
   ++token_stream->pos;
   return(token);
}

/**************************************************************/
/*  FUNCTION   : compare_token 
 *  INPUTS     : token1, token2 
 *  OUTPUTS    : boolean	 
 *  DESCRIPTION: compare two tokens.  Tokens of the same type are
 * 		 equal unless the token is a keyword. Then the
 * 		 keywords must be the same.  
 */ 

compare_token(token1,token2) 
TOKEN token1, token2;
{
   if (!strcmp(token1.first_element,token2.first_element))
      {
   	if (!strcmp(token1.first_element,"keyword"))
           {
             if (strcmp(token1.second_element,token2.second_element))
                {
                   return(TRUE);
                }
  	   }
      }
   else
	return(TRUE);

   return(FALSE);   
}
 
/**************************************************************/
/*  FUNCTION   : print_token 
 *  INPUTS     : token_string 
 *  OUTPUTS    : token is printed to standard output 
 *  DESCRIPTION: a given token is printed to the screen.
 *		 Format:
 *		 	for symbols  <type>.
 *			for all else <type>,<value> 
 *		 all values except numeric values are enclosed in " ".
 */

print_token(token_string)
TOKEN token_string;
{
   if (token_string.first_element[0] != NULL)
      {  
	  printf("%-s",token_string.first_element);   
      }
   if ((token_string.second_element[0] != NULL)&&
	(strcmp(token_string.first_element,"numeric"))) 
      {  
          printf(", \t\"%-s\".\n",token_string.second_element);
      }
   else if ((token_string.first_element[0] != NULL) &&  
		(token_string.second_element[0] != NULL))
      {  
        printf(", \t%-s.\n",token_string.second_element);
      }
   else
          printf(".\n");
}

/**************************************************************/
/*  FUNCTION   : is_eof_token 
 *  INPUTS     : token 
 *  OUTPUTS    : boolean 
 *  DESCRIPTION: This functions checks to see if the end of the
 *  		 file token was encounted. 
 */ 
is_eof_token(token) 
TOKEN token;
{
   if (!strcmp(token.first_element,"eof"))
	return (TRUE);   
   else
	return (FALSE);   

}

