/*************************************************************************
	This module is for operations on "lines."  Lines consist of
	an ordered set of words, and each word is an ordered set of
	characters.  Words in lines are separated by whitespace 
	characters, such as tabs, spaces, etc.  The operations 
	implemented thus far are:
		
		num_of_words	-  counts words in a line
		string_to_line	-  converts string to line
		append_to_line	-  appends string to line
		line_to_string	-  converts line to string
		line_cmp	-  compares lines
		circular_shift	-  circularly shifts lines
		dump_line	-  prints line to stderr for debugging

*************************************************************************/

#include "line.h"


/****************************************************************************
Function:	error
Parameters:	one "C type" string
Returns:	void
Purpose:	This function prints out message to stderr, if not NULL 
		and then exits the program.
****************************************************************************/

static void error (char *message)
{
  if (message != NULL)
    fprintf (stderr, "\n%s\n\n", message);
  exit (1);
}


 
/*************************************************************************
Function:	num_of_words
Parameters:	one of type line
Returns:	An int representing the number of words in the
		line
Purpose:	The purpose of this function is to count the
		number of words contained in a variable of
		type line.
Note:		Variable(s) passed in is(are) not modified
*************************************************************************/

int num_of_words (line *any_old_line)
{
  int count;
  line *temp_line;

  if (any_old_line == NULL)
    error ("line.c, num_of_words:  Bad parameter");
  if (strcmp (any_old_line->word, DELIMITERS) == 0)
    return (0);
  count = 0;
  temp_line = any_old_line;
  while (temp_line != NULL)
  {
    count = count + 1;
    temp_line = temp_line->next;
  }
  return (count);
}


 
/*************************************************************************
Function:	string_to_line
Parameters:	One null terminated "C type" string
Returns:	A variable of type line containing the string
		that was passed in.
Purpose:	Takes a string and returns a variable of type
		line.  This function also removes all white-
		space characters from the string before con-
		version to a line.
Note:		Variable(s) passed in is(are) not modified
*************************************************************************/

line *string_to_line (char *string)
{
  char *temp_str, *word;
  line *new_word, *prev_word, *temp_word;
  
  if (string == NULL) error ("line, string_to_line:  Bad parameter");
  temp_str = (char *) malloc ((strlen (string) + 1) * sizeof (char));
  if (temp_str == NULL) error ("line, string_to_line:  Allocation error");
  strcpy (temp_str, string);
  word = strtok (temp_str, DELIMITERS);
  new_word = (line *) malloc (sizeof (line));
  if (new_word == NULL) error ("line, string_to_line:  Allocation error");
  if (word != NULL)
    new_word->word = (char *) malloc ((strlen (word) + 1) * sizeof (char));
  else
    new_word->word = (char *) malloc (sizeof DELIMITERS);
  if (new_word->word == NULL) error ("line, string_to_line:  Allocation error");
  new_word->next = NULL;
  prev_word = new_word;
  if (word != NULL)
    strcpy (new_word->word, word);
  else
  {
    strcpy (new_word->word, DELIMITERS);
    return (new_word);
  }
  while ((word = strtok (NULL, DELIMITERS)) != NULL)
  {
    temp_word = (line *) malloc (sizeof (line));
    if (temp_word == NULL) error ("line, string_to_line:  Allocation error");
    temp_word->word = (char *) malloc ((strlen (word) + 1) * sizeof (char));
    if (temp_word->word == NULL) error ("line, string_to_line:  Allocation error");
    temp_word->next = NULL;
    strcpy (temp_word->word, word);
    prev_word->next = temp_word;
    prev_word = temp_word;
  }
  return (new_word);
}

 

/*************************************************************************
	Function:       	append_to_line
	Parameters:     	One of type line and one "C type" string
	Returns:       	A variable of type line containing the original
			line with the string appended to it.
	Purpose:        	This function converts a string to type line
			and appends it to a previously created line.
	Note:           	Variable(s) passed in may be modified
*************************************************************************/

line *append_to_line (line *any_old_line, char *string)
{
  char *temp_str, *word;
  line *new_word, *temp_word, *prev_word;

  if ((any_old_line == NULL) || (string == NULL))
    error ("line, append_to_line:  Bad parameter");
  if (strcmp (any_old_line->word, DELIMITERS) == 0)
    return (string_to_line (string));
  temp_str = (char *) malloc ((strlen (string) + 1) * sizeof (char));
  if (temp_str == NULL)
    error ("line, append_to_line:  Memory allocation error");
  strcpy (temp_str, string);
  word = strtok (temp_str, DELIMITERS);
  if (word == NULL) return (any_old_line);
  new_word = (line *) malloc (sizeof (line));
  new_word->word = (char *) malloc ((strlen (word) + 1) * sizeof (char));
  new_word->next = NULL;
  temp_word = any_old_line;
  while (temp_word->next != NULL) temp_word = temp_word->next;
  temp_word->next = new_word;
  prev_word = new_word;
  strcpy (new_word->word, word);
  while ((word = strtok (NULL, DELIMITERS)) != NULL)
  {
    temp_word = (line *) malloc (sizeof (line));
    temp_word->word = (char *) malloc ((strlen (word) + 1) * sizeof (char));
    temp_word->next = NULL;
    strcpy (temp_word->word, word);
    prev_word->next = temp_word;
    prev_word = temp_word;
  }
  return (any_old_line);
}


 
/*************************************************************************
	Function:       	line_to_string
	Parameters:     	One of type line
	Returns:        	One null terminated "C type" string containing
			the line that was passed in.
	Purpose:        	The function is used to convert variables of
			type line back to the more widely used string
			type.
	Note:      	Variable(s) passed in is(are) not modified
*************************************************************************/

char *line_to_string (line *any_old_line)
{
  char *temp_str;
  line *temp_line;

  if (any_old_line == NULL)
    error ("line, line_to_string:  Bad parameter");
  if (strcmp (any_old_line->word, DELIMITERS) == 0)
  {
    temp_str = (char *) malloc (sizeof (char));
    temp_str[0] = '\0';
    return (temp_str);
  }
  temp_line = any_old_line;
  temp_str = (char *) malloc ((strlen (temp_line->word) + 1) * sizeof (char));
  if (temp_str == NULL)
    error ("line, line_to_string:  Memory allocation error");
  strcpy (temp_str, temp_line->word);
  temp_line = temp_line->next;
  while (temp_line != NULL)
  {
    temp_str = (char *) realloc 
      (temp_str, ((strlen (temp_line->word) + 2) + strlen (temp_str)));
    if (temp_str == NULL)
      error ("line, line_to_string:  Memory allocation error");
    strcat (temp_str, " ");
    strcat (temp_str, temp_line->word);
    temp_line = temp_line->next;
  }
  return (temp_str);
}


 
/*************************************************************************
	Function:       	line_cmp
        	Parameters:   	Two of type line
        	Returns:        	An int greater than, equal to, or less than 0,
                        		according as l1 is lexicographically greater 
                        		than, equal to, or less than l2.
        	Purpose:        	This function works similarly to if not 
                        		identically to strcmp and serves the same purpose.
        	Note:           	Variable(s) passed in is(are) not modified
*************************************************************************/

int line_cmp (line *any_old_line1, line *any_old_line2)
{
  char *string1, *string2;
  int result;
  
  if ((any_old_line1 == NULL) || (any_old_line2 == NULL))
    error ("line, line_cmp:  Bad parameter");
  if (strcmp (any_old_line1->word, DELIMITERS) == 0)
    string1= "\0";
  else
    string1 = line_to_string (any_old_line1);
  if (strcmp (any_old_line2->word, DELIMITERS) == 0)
    string2= "\0";
  else
    string2 = line_to_string (any_old_line2);
  result = strcmp (string1, string2);
  if (result == 0) return (result);
  else return (result / abs (result));
}


 
/*************************************************************************
        	Function:       	circular_shift
        	Parameters:     	One of type line
        	Returns:        	A variable of type line
       	Purpose:        	This function is used to circular shift a line.
                        		This is done by moving the first word from the
                        		beginning of the line to the end of it.
        	Note:           	Variable(s) passed in is(are) not modified
*************************************************************************/

line *circular_shift (line *any_old_line)
{
  line *first_word, *temp_line, *temp_word;
  
  if (any_old_line == NULL)
    error ("line, circular_shift:  Bad parameter");
  if (strcmp (any_old_line->word, DELIMITERS) == 0)
    return (any_old_line);
  if (any_old_line->next == NULL) return (any_old_line);
  first_word = any_old_line;
  temp_line = any_old_line->next;
  temp_word = any_old_line->next;
  first_word->next = NULL;
  while (temp_word->next != NULL) temp_word = temp_word->next;
  temp_word->next = first_word;
  return (temp_line);
}  


/*************************************************************************
        	Function:       	dump_line
        	Parameters:     	One "C type" string and one of type line
        	Returns:        	void
        	Purpose:        	This function is useful for debugging.
        	Note:           	Variable(s) passed in is(are) not modified
*************************************************************************/

void dump_line (char *message, line *any_old_line)
{
  if ((message == NULL) || (any_old_line == NULL))
    error ("line, dump_line:  Bad parameter");
  fprintf
    (stderr, "\n%s:  \"%s\"\n", message, line_to_string (any_old_line));
}
