#ifndef _LINE_H
#define _LINE_H


/*************************************************************************
	This module is for operations on "lines."  Lines consist of
	an ordered set of words, and each word is an ordered set of
	characters.  Words in lines are separated by whitespace 
	characters, such as tabs, spaces, etc.  The operations 
	implemented thus far are:
		
		num_of_words	-  counts words in a line
		string_to_line	-  converts string to line
		append_to_line	-  appends string to line
		line_to_string	-  converts line to string
		line_cmp	-  compares lines
		circular_shift	-  circularly shifts lines
		dump_line	-  prints line to stderr for debugging

*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define DELIMITERS "\n\t "

typedef struct line_node
{
  char *word;
  struct line_node *next;
} line;


/*************************************************************************
Function:	num_of_words
Parameters:	one of type line
Returns:	An int representing the number of words in the
		line
Purpose:	The purpose of this function is to count the
		number of words contained in a variable of
		type line.
Note:		Variable(s) passed in is(are) not modified
*************************************************************************/

int num_of_words (line *any_old_line);


 
/*************************************************************************
Function:	string_to_line
Parameters:	One null terminated "C type" string
Returns:	A variable of type line containing the string
		that was passed in.
Purpose:	Takes a string and returns a variable of type
		line.  This function also removes all white-
		space characters from the string before con-
		version to a line.
Note:		Variable(s) passed in is(are) not modified
*************************************************************************/

line *string_to_line (char *string);


/*************************************************************************
	Function:       	append_to_line
	Parameters:     	One of type line and one "C type" string
	Returns:       	A variable of type line containing the original
			line with the string appended to it.
	Purpose:        	This function converts a string to type line
			and appends it to a previously created line.
	Note:           	Variable(s) passed in may be modified
*************************************************************************/

line *append_to_line (line *any_old_line, char *string);


/*************************************************************************
	Function:       	line_to_string
	Parameters:     	One of type line
	Returns:        	One null terminated "C type" string containing
			the line that was passed in.
	Purpose:        	The function is used to convert variables of
			type line back to the more widely used string
			type.
	Note:      	Variable(s) passed in is(are) not modified
*************************************************************************/

char *line_to_string (line *any_old_line);


/*************************************************************************
	Function:       	line_cmp
        	Parameters:   	Two of type line
        	Returns:        	An int greater than, equal to, or less than 0,
                        		according as l1 is lexicographically greater 
                        		than, equal to, or less than l2.
        	Purpose:        	This function works similarly to if not 
                        		identically to strcmp and serves the same purpose.
        	Note:           	Variable(s) passed in is(are) not modified
*************************************************************************/

int line_cmp (line *any_old_line1, line *any_old_line2);
 


/*************************************************************************
        	Function:       	circular_shift
        	Parameters:     	One of type line
        	Returns:        	A variable of type line
       	Purpose:        	This function is used to circular shift a line.
                        		This is done by moving the first word from the
                        		beginning of the line to the end of it.
        	Note:           	Variable(s) passed in is(are) not modified
*************************************************************************/

line *circular_shift (line *any_old_line);


/*************************************************************************
        	Function:       	dump_line
        	Parameters:     	One "C type" string and one of type line
        	Returns:        	void
        	Purpose:        	This function is useful for debugging.
        	Note:           	Variable(s) passed in is(are) not modified
*************************************************************************/

void dump_line (char *message, line *any_old_line);

#endif

