/****************************************************************************
Module:	line_storage

This module implements a type called line_storage and some operations that
can be performed on this type.  The currently supported operations are:

	empty_line_storage	-	creates empty line storage
	circulate_and_add_line	-	adds line and all circular shifts to
					line storage
	get_line			-	gets an arbitrary line from line storage
	num_of_lines		-	counts the lines in line storage
	alphabetize		-	sorts the lines in alphabetical order
****************************************************************************/

#include "line_storage.h"


/****************************************************************************
Function:	error
Parameters:	one "C type" string
Returns:	void
Purpose:	This function prints out message to stderr, if not NULL 
		and then exits the program.
****************************************************************************/

static void error (char *message)
{
  if (message != NULL)
    fprintf (stderr, "\n%s\n\n", message);
  exit (1);
}


 
/****************************************************************************
Function:	make_node
Parameters:	none
Returns:	a pointer to type node
	Purpose:	This is a private function used to create an empty node, used
			in line_storage.
****************************************************************************/

node *make_node ()
{
  node *new_node;
  
  new_node = (node *) malloc (sizeof (node));
  if (new_node == NULL)
    error ("line_storage, make_node:  Memory allocation error");
  new_node->contents = NULL;
  new_node->next = NULL;
  return (new_node);
}


/****************************************************************************
Function:	empty_line_storage
Parameters:	none
Returns:	a pointer to type line_storage
Purpose:	This function creates and returns a variable of type line_storage.
****************************************************************************/

line_storage *empty_line_storage ()
{
  line_storage *new_storage;

  new_storage = (line_storage *) malloc (sizeof (line_storage));
  if (new_storage == NULL)
    error ("line_storage, empt_line_storage:  Memory allocation error");
  new_storage->head = NULL;
  new_storage->num_of_lines = 0;
  return (new_storage);
}


 
/****************************************************************************
Function:	circulate_and_add_line
Parameters:	one pointer to type line_storage and one "C type" string
Returns:	a pointer to type line_storage
Purpose:	This function takes string, converts it to a line and then
		adds all of the circular shifts of that line to storage.
Note:		Variable(s) passed in may be modified after function call.
****************************************************************************/

line_storage *circulate_and_add_line (line_storage *storage, char *string)
{
  int count, words;
  line *prev_line, *temp_line;
  node *prev_node, *curr_node, *new_node;

  if ((storage == NULL) || (string == NULL))
    error ("line_storage, circulate_and_add_line:  Bad parameter");
  prev_line = string_to_line (string);
  if (prev_line == NULL)
    error ("line_storage, circulate_and_add_line:  Error creating line");
  words = num_of_words (prev_line);
  for (count = 1; count <= words; count++)
  {
    prev_node = curr_node = storage->head;
    temp_line = circular_shift (prev_line);
    if (temp_line == NULL)
      error ("line_storage, circulate_and_add_line:  Error shifting line");
    while ((curr_node != NULL) && 
	   (line_cmp (curr_node->contents, temp_line) < 1))
    {
      prev_node = curr_node;
      curr_node = curr_node->next;
    }
    new_node = make_node ();
    new_node->next = curr_node;
    new_node->contents = temp_line;
    if ((storage->num_of_lines == 0) || (curr_node == storage->head))
      storage->head = new_node;
    else
      prev_node->next = new_node;
    storage->num_of_lines++;
    prev_line = string_to_line (line_to_string (temp_line));
  }    
}


 
/****************************************************************************
Function:	get_line
Parameters:	one pointer to type line_storage and one integer
Returns:	a "C type" string
Purpose:	This function returns the line specified by position in 
		storage as a string.  Program will exit if position is
		out of range.
****************************************************************************/

char *get_line (line_storage *storage, int position)
{
  int count;
  node *curr_node;

  if ((storage == NULL) || (position > storage->num_of_lines))
    error ("line_storage, get_line:  Bad Parameter");
  curr_node = storage->head;
  for (count = 1; count < position; count++)
    curr_node = curr_node->next;
  return (line_to_string (curr_node->contents));
}


/****************************************************************************
Function:	num_of_lines
Parameters:	one pointer to type line_storage
Returns:	an integer
Purpose:	This function returns the number of lines in storage.
****************************************************************************/

int num_of_lines (line_storage *storage)
{
  if (storage == NULL)
    error ("line_storage, num_lines:  Bad parameter");
  return (storage->num_of_lines);
}


/****************************************************************************
Function:	alphabetize
Parameters:	one pointer to type line_storage
Returns:	a pointer to type line_storage
Purpose:	This function if supposed to sort the lines in storage
		alphabetically, but in this implementation the lines are
		always sorted.
****************************************************************************/

line_storage *alphabetize (line_storage *storage)
{
  if (storage == NULL)
    error ("line_storage, alphabetize:  Bad parameter");
  return (storage);
}

 

/****************************************************************************
Function:	dump_line_storage
Parameters:	one "C type" string and one pointer to type line_storage
Returns:	void
Purpose:	This function prints message to stderr and then proceeds to
		"dump" the entire contents of storage.
****************************************************************************/

void dump_line_storage (char *message, line_storage *storage)
{
  char line_num[16];
  int count;
  node *temp_node;

  if ((message == NULL) || (storage == NULL))
    error ("line_storage, dump_line_storage:  Bad parameter");
  fprintf (stderr, "%s\n", message);
  if (storage->num_of_lines == 0)
    fprintf (stderr, "<LINE STORAGE IS EMPTY>\n");
  else
  {
    count = 1;
    temp_node = storage->head;
    do
    {
      sprintf (line_num, "line #%d ", count);
      dump_line (line_num, temp_node->contents);
      temp_node = temp_node->next;
      count++;
    } while (temp_node != NULL);
  }
}

